\documentclass[a4paper,12pt]{article}

\usepackage[utf8x]{inputenc}
\usepackage[top=2.5cm, bottom=2.5cm, left=3.5cm, right=2.5cm]{geometry}
\usepackage{setspace}
\usepackage{graphicx}
\usepackage{float}
\usepackage{wrapfig}
\usepackage{multirow}

\usepackage[magyar]{babel}

\usepackage{url}
\usepackage{cite}

\setcounter{secnumdepth}{5}
\setcounter{tocdepth}{5}

%include agda.fmt

%format λ  = "\lambda"
%format σ  = "\sigma"

\long\def\ignore#1{}

\spacing{1.5}

\begin{document}

\ignore{
\begin{code}
module Mastermind.Documentation where
\end{code}
}

\begin{titlepage}

\title{Interaktív program megvalósítása Agdában}
\author{Hegedűs Dénes}
\date{2013}

\begin{tabular}{ l c }
\multirow{4}{*}{\includegraphics[width=120px]{pic/elte_cimer/nagy_cimer.jpeg}}
 & Eötvös Loránd Tudományegyetem \\
 & Informatikai Kar \\[0.5cm]
 & Programozási Nyelvek \\
 & és Fordítóprogramok Tanszék \\
\end{tabular}\\

\line(1,0){400} \\[2cm]

\begin{center}
\textsc{\large Interaktív program megvalósítása Agdában} \\[2cm]
\end{center}

\begin{minipage}{0.5\textwidth}
  \begin{flushleft} \large
    \emph{Témavezetők:}\\
    \textbf{Diviánszky Péter} \\
    tanársegéd \\[0.5cm]
    \textbf{Páli Gábor János} \\
    tanársegéd
  \end{flushleft}
\end{minipage}
\begin{minipage}{0.5\textwidth}
  \begin{flushleft} \large
    \emph{Szerző:} \\
    \textbf{Hegedűs Dénes} \\
    Programtervező Informatikus BSc
  \end{flushleft}
\end{minipage}

\vfill

\begin{center}
Budapest, 2013
\end{center}


\end{titlepage}


\thispagestyle{empty}
\tableofcontents 
\newpage


\section{Bevezetés}

\subsection{A feladat megfogalmazása}

A szakdolgozat célja egy olyan egyszerű és jól használható interaktív program megvalósítása, ami bizonyos garanciákkal is rendelkezik. Ilyen elvárások lehetnek például, hogy a program ne kerülhessen végtelen ciklusba, vagy nem várt állapotba, ezen kívül bizonyos függvények helyes működését is szeretnénk garantálni. \\

\subsection{Agda}

Ezeknek az elvárásoknak a megvalósításához az Agda nevű tisztán funkcionális nyelvet használjuk fel, ami integrált tételbizonyító rendszerrel és erős típusrendszerrel rendelkezik, amelyek segítségével fordítás időben tudjuk biztosítani a következőket:

\begin{enumerate}
  \item A program nem kerülhet végtelen ciklusba.
  \item A programban nem léphet fel a programozó hibájából adódó futásidejű hiba.
  \item A tesztelés része a programnak, ami szintén fordításidőben történik, így kizáró- lag olyan program állítható elő a fordítóval, amire a tesztesetek az elvárások szerint futnak le.
\end{enumerate}

\subsection{Funkcionális Reaktív Programozás}

A program elkészítése során felhasználjuk a Funkcionális Reaktív Programozást (Functional Reactive Programming, FRP), aminek segítségével tisztán funkcionálisan, időtől függő folytonos függvények segítségével írhatunk le interaktív programokat. Ennek a modellnek az alapja a Funkcionális Reaktív Animáció (Fran)\cite{ElliottHudak97:Fran} amivel interaktív animációk készíthetők el. \\

Ennek megvalósításához az Agda nyelvben írt agda-frp-js függvénykönyvtárat használjuk fel, ami ECMAScript kimenetű, így böngészőben futtatható programot állít elő. \\

\subsection{A programozási feladat}

A megvalósítandó program a Mastermind nevű klasszikus táblajáték virtuális változata. \\

Ez egy kétfős kód-feltörő játék, amiben az egyik játékos a kód kitalálója, a másik pedig a kódfeltörő. A kód kitalálója megadott színek (6 féle) közül kiválaszt egy 4 elemű mintát, amit aztán a kódfeltörő minél kevesebb lépésben próbál megfejteni. A minta tartalmazhatja ugyanazt a színt többször is és a játék során a kódfeltörő számára rejtve marad. A próbálkozások számára felső korlát van (alapesetben 12), amit ha túllépünk, akkor a kód kitalálója nyert, ha pedig sikerül megfejteni, akkor a kódfeltörő. \\

A megvalósított programban mindig a számítógép a kód kitalálója, a játékos pedig a kód feltörője.


\newpage


\section{Felhasználói dokumentáció}

\subsection{Bevezetés}

A felhasználói dokumentáció tartalmazza a program futtatásához és használatához szükséges információkat, első részében bemutatjuk a játék logikai szerkezetét, második részében pedig a felhasznált keretrendszert és a rendszerkövetelményeket.

\subsection{Felhasznált módszerek}

A program elkészítéséhez az Agda\cite{website:agda-wiki} nevű funkcionális nyelvet és az agda-frp-js\cite{website:agda-frp-js} függvénykönyvtárat használtuk fel. Ez a függvénykönyvtár ECMAScript\cite{ECMA-262} kimenetet hoz létre, ami aztán kliens-oldali webes alkalmazásként webböngészőkben futtatható. \\

A felhasznált módszerek részletes leírásához lásd a Fejlesztői dokumentáció 3.3. fejezetét.

\subsubsection{Agda}

Az Agda\cite{website:agda-wiki} egy olyan funkcionális nyelv, ami integrált tételbizonyító rendszerrel rendelkezik, mely segítségével már fordítási időben tudjuk garantálni, hogy a program nem állhat le hibával, továbbá végtelen ciklusba se kerülhet. Emellett további állítások és megkötések is könnyen leírhatók benne, amikkel tetszőleges hibás állapotokat is elkerülhetünk, erős típusrendszere pedig kiszűri a típushibákat. \\

A program tesztelése a tételbizonyító rendszerrel már fordításidőben megvalósít- ható, ezzel biztosítva, hogy csak olyan programot tudunk lefordítani, amire a tesztese- tek megfelelően működnek.

\subsubsection{agda-frp-js}

Az agda-frp-js\cite{website:agda-frp-js} egy nyílt forráskódú függvénykönyvtár, aminek segítségével reaktív funkcionális programok írhatók le. A reaktív funkcionális programozás (Functional Reactive Programming, FRP) egy egyre inkább terjedő módszer interaktív programok készítésére, amellyel időtől függő függvényekkel írhatunk le interaktív programokat.

\subsection{Telepítő}

A programhoz Windows operációs rendszerhez telepítő tartozik. Ennek segítségével feltelepíthetjük a szoftvert egy tetszőleges mappába, amihez Start menübeli parancsikon és uninstaller is létrehozásra kerül. A telepítő 3 lépésből áll, az első ablakban a programról látható információ, az Elfogadom gombbal tudunk tovább lépni.

\begin{figure}[H]
  \begin{center}
  \includegraphics[scale=0.75]{pic/install1.png}
  \caption{Telepítő}
  \end{center}
\end{figure}

Ezután megkérdi az installer a telepítés helyét, ezután a Telepítés gombra kattintva elindul a telepítés, a Start menübeli parancsikon létrehozása és a Registrybe való felvétel.

\begin{figure}[H]
  \begin{center}
  \includegraphics[scale=0.75]{pic/install2.png}
  \caption{Telepítő - Telepítés helye}
  \end{center}
\end{figure}

A telepítés végeztével megtekinthetjük a felmásolt fájlokat vagy bezárhatjuk a telepítőt:

\begin{figure}[H]
  \begin{center}
  \includegraphics[scale=0.75]{pic/install3.png}
  \caption{Telepítő - Vége}
  \end{center}
\end{figure}

\subsection{Forrásból való fordítás}

A forrásból való fordításhoz szükséges az Agda 2.3.0.1-es verziója, és az agda-frp-js csomag lefordított változata, ennek megadásával Makefile segítségével (GNU Make 3.81) tudunk fordítani a forrás gyökérkönyvtárában kiadott paranccsal: \\

\begin{center}
\texttt{make FRP\_DIR=<agda-frp-js helye>}
\end{center}

Ez a parancs a dist mappába fordítja és másolja a kellő fájlokat, ezután a program futtatható is.

\subsection{Futtatás}

Az lefordított kódhoz egy egyszerű HTML keretet hozunk létre, ami a \texttt{main.html} fájlban található, a stílusát pedig a \texttt{mastermind.css} nevű stíluslapban írtuk le. \\

A játék a <Telepítés helye>/$main.html$ megnyitásával indítható. \\

\begin{figure}[H]
  \begin{center}
  \includegraphics[scale=0.5]{pic/fut.png}
  \caption{A program futás közben}
  \end{center}
\end{figure}

Windows operációs rendszeren a telepítővel feltelepített program a Start menüben létrehozott parancsikonnal is indítható:

\begin{figure}[H]
  \begin{center}
  \includegraphics{pic/start_menu.png}
  \caption{Start menü}
  \end{center}
\end{figure}

\subsection{A program működése}

A program kinézetre világos hátterű, ezen belül a sötétebb keretben helyezkedik el az interaktív része, ami több részből áll:

\begin{enumerate}
  \item Felül, a keret körülbelül 2/3-át elfoglaló rész mutatja a korábbi próbálkozásokat, ez 12 sorból áll.
  \item Alatta a tipp látható, ami 1 sor és szélesebb a történetnél.
  \item A tipp alatt, miután véget ért a játék egy felirat lesz látható, ami jelzi, hogy nyertünk-e vagy vesztettünk.
  \item Legalul a játékban való lépést és egyéb interakciókat megvalósító gombok helyezkednek el.
\end{enumerate}

\begin{figure}[H]
  \begin{center}
  \includegraphics[scale=0.5]{pic/order.png}
  \caption{Elrendezés}
  \end{center}
\end{figure}

\subsubsection{Korábbi próbálkozások}

A korábbi próbálkozások listája kezdetben üres sorokat tartalmaz, ami a játék során felülről lefelé töltődik. Egy sor bal oldalán a tipp látható, mellette pedig sorrendtől függetlenül a fekete pöttyök olyan elemeket jelölnek, amik a tippben és a megoldásban színre és helyre is megegyeznek, a fehérek pedig olyan színeket, amik a tippben és a megoldásban is megtalálhatók, viszont más pozíción.

\begin{figure}[H]
  \begin{center}
  \includegraphics[scale=0.5]{pic/history.png}
  \caption{A történet}
  \end{center}
\end{figure}

\subsubsection{A tipp}

A korábbi próbálkozások alatt helyezkedik el az aktuális tipp, aminek változtatható mezőjét belül fehér kör jelöli.

\begin{figure}[H]
  \begin{center}
  \includegraphics[scale=0.5]{pic/guess.png}
  \caption{A tipp}
  \end{center}
\end{figure}

\subsubsection{A gombok}

A tipp után következnek a gombok:

\begin{figure}[H]
  \begin{center}
  \includegraphics[scale=0.5]{pic/buttons.png}
  \caption{A gombok}
  \end{center}
\end{figure}

\begin{enumerate}
  \item A színes gombokkal változtathatjuk az aktuálisnak megjelölt mező színét.
  \item A '$<$' és '$>$' feliratúakkal tudjuk léptetni az aktuális mezőt.
  \item A 'Check' gombbal tudjuk ellenőriztetni a tippet.
  \item A 'Clear' gomb visszaállítja a tipp mintáját alapértelmezett (kék, kék, kék, kék) mintára.
  \item A 'Restart' gomb pedig új játékot kezd.
\end{enumerate}

\subsubsection{A játék vége}

A játék kétféleképp érhet véget, nyerés esetén ,,You won! :)'' felirat látható, vesztés esetén pedig ,,You lost! :( Solution: [abcd]'', ahol [abcd] jelöli a megoldásban lévő színek indexét.

\subsection{Rendszerkövetelmények}

A program tetszőleges operációs rendszeren, HTML5-t támogató, JavaScriptet futtatni képes webböngészőben futtatható, a következő operációs rendszerekben és böngészőkben teszteltem:

\begin{itemize}
  \item GNU/Linux (Debian)
    \begin{itemize}
      \item Mozilla Firefox 20.0
    \end{itemize}
  \item Windows
    \begin{itemize}
      \item Mozilla Firefox 20.0.1
      \item Google Chrome 26.0.1410.64
    \end{itemize}
\end{itemize}

Internet Explorer esetén a böngésző JavaScript motorjának hiányosságai miatt a program nem futtatható.


\newpage


\section{Fejlesztői dokumentáció}

\subsection{Bevezetés}

Fő célunk egy letisztult, rövid és egyszerű program megvalósítása, ami a felhasznált eszközökből kifolyólag olyan plusz garanciákat is ad, ami más nyelveken nem, vagy csak nehezen valósítható meg. \\

Ehhez felhasználjuk az Agda nevű tisztán funkcionális nyelvet és a hozzá megírt agda-frp-js csomagot, aminek segítségével ECMAScript kimenetű, böngészőben futtatható interaktív programot állítunk elő. \\

Emellett a fordítás során szeretnénk tesztelni a program működését, így további függvényekként egy külön modulban teszteseteket és további állításokat fogalmazunk meg a programról, amiket csak akkor fogad el az Agda típusellenőrzője, ha teljesülnek, ezután ezeket nem használjuk fel a program futása során, azonban lefordulásuk előfeltétele lesz annak, hogy a program maga is leforduljon. \\

A megvalósított program a Mastermind nevű játék, amivel az interakció gombokon keresztül fog történni.

\subsection{Elvárások}

Legfőbb elvárásunk a programmal szemben, hogy típusai minél speciálisabbak legye- nek, így a rajtuk alkalmazott függvények nem vihetik át a programot hibás vagy nem várt állapotba, továbbá szeretnénk bizonyításokkal alátámasztani a függvények helyes működését. \\

A programot fontos, hogy több, kisebb modulra bontsuk, amik jól különválasztha- tóak és egyfajta hierarchia építhető fel rajtuk, így a különböző részei külön tesztelhetők. Mivel a megvalósítandó program interaktív, így célszerű különválasztanunk a megjelenítést a játék logikájától, ahol a nézet importálja a logikát. Emellett a program tesztelését is szeretnénk egy olyan modulban megadni, aminek a fordítása közben értékelődnek ki a tesztesetek. Ez a modul a modell és a nézet függvényeit is felhasználhatja, de nem állíthat elő használható kódot. \\

A megvalósítás során hatékonysági szempontokat is figyelembe kell vennünk, mivel tisztán funkcionális programok esetén alkalmazott többszörösen egymásba ágyazott rekurziók lassú futást és nagy memóriaigényt eredményezhetnek.

\subsection{Felhasznált módszerek}

Ebben a részben bemutatjuk az Agda főbb nyelvi elemeit és a felhasznált függvénykönyvtár alacsony szintű függvényeit.

\subsubsection{Agda}

\ignore{
\begin{code}
module Examples where
  open import FRP.JS.Nat hiding (_≟_ ; _<_)
\end{code}
}

Az Agda egy függő típusos tisztán funkcionális nyelv\cite{norell:thesis}, amiben Unicode\cite{Unicode:2012:USV} karaktereket is felhasználhatunk. A függvények megadása hasonló a matematikában alkalmazotthoz, a típusokat szinte minden esetben ki kell írnunk, a függvény paramétereinél a zárójelek pedig elhagyhatók, így például egy függvény, ami összead két természetes számot hasonlóképp néz ki:

\begin{code}
  f : ℕ → ℕ → ℕ
  f x y = x + y
\end{code}

Típusokat konstruktoraik megadásával írhatunk le, amik tetszőleges paraméterszámúak lehetnek, például egy két elemű típus megadása:

\begin{code}
  data KetElem : Set where
    elem₁ : KetElem
    elem₂ : KetElem
\end{code}

A típusok lehetnek paraméteresek és függhetnek konkrét értékektől (függő típus) is, például az \texttt{A} típusú elemeket tartalmazó \texttt{n} hosszú vektorok egy definíciója:

\begin{code}
  data Vec (A : Set) : ℕ → Set where
    []  : Vec A 0
    _∷_ : ∀ {n : ℕ} (x : A) (xs : Vec A n) → Vec A (n + 1)

  -- üres vektor ([ ]) 0 hosszú
  -- x-et egy n hosszú xs vektor elé fűzve (x ∷ xs) n + 1 hosszú vektor lesz
\end{code}

Itt látható egy \texttt{\{n : ℕ\}} jelölés is, ez a kapcsos zárójel rejtett paramétert jelöl, amit a függvény törzsekben nem kell kiírnunk paraméterként akkor, amikor a fordító ki tudja azt következtetni. \\

Az Agda nyelvben továbbá kizárólag totális függvények adhatók meg, azaz például a funkcionális programozásból ismert \texttt{head} függvény, ami egy tetszőleges vektor első elemét adná vissza, nem adható meg:

\begin{code}
  head : ∀ {n : ℕ} {A : Set} → Vec A n → A
  head (x ∷ xs)  = x
  head []        = {!!} -- hiba
\end{code}

Itt üres vektor esetén egy olyan elemet kellene megadnunk, ami tetszőleges \texttt{A} halmaznak eleme, azonban ilyen Agdában nem lehetséges, így ilyen típusú függvényt nem tudunk megadni. \\

Ennek a tulajdonságnak köszönhetően nem tud a program ,,elszállni'' és nem tudunk a program futása során kivételeket sem dobni, hanem minden függvénynek át kell vinnie a programot egy másik állapotba, ezzel sok hibás esetet kiszűrve. \\

A nyelv rendelkezik továbbá mixfix operátorokkal, névtelen (lambda) függvények- kel és kötési erősség megadásával is, ezekre egy-egy példa:

\begin{code}
  -- függvénykompozíció
  _∘_ : ∀ {A B C : Set} (f : B → C) → (g : A → B) → (A → C)
  f ∘ g = λ x → f (g x)

  -- balról zárójeleződő, 3-as precedenciájú operátor
  infixl 3 _∘_
\end{code}

Megadhatók a nyelvben olyan függvények és értékek is, amiknek típusuk van, de Agda-beli definíció nem tartozik hozzájuk, ezeket a \texttt{postulate} kulcsszó segítségével tudjuk bevezetni. Ennek használata kerülendő, mivel így tetszőleges típushoz tudunk új elemeket megadni, amivel kijátszható a típusellenőrzés.

\subsubsection{Funkcionális Reaktív Programozás}

\ignore{
\begin{code}
module Reactive where
  open import FRP.JS.Nat hiding (_≟_ ; _<_ ; _≤_ )
  open import FRP.JS.Time.Core
  open import FRP.JS.Delay using ( _≤_ )
  open import FRP.JS.Bool

  open import Mastermind.Auxiliary hiding (_<_)
\end{code}
}

A felhasznált agda-frp-js függvénykönyvtár a Funkcionális Reaktív Programozás (Functional Reactive Programming - FRP) egy modelljének\cite{Jef12:LTLTypesFRP} implementációját tartalmazza. Ebben a modellben időtől függő mellékhatásmentes függvények segítségével adható meg a programmal való interakció. \\

A függvénykönyvtár tartalmazza a modellt, az interakciót segítő magasabb szintű függvényeket és az ECMAScript kimenetet előállító részeket is, moduljainak elnevezése \texttt{FRP.JS.}\emph{modulnév}. \\

Ebben a modellben a legfőbb típusok a reaktív típusok, amik idő szerint változnak, ennek definícióját a könyvtárban (module FRP.JS.RSet) egyszerű típusszinonimaként adták meg:

\begin{code}
  RSet : Set₁
  RSet = Time → Set
\end{code}

Azaz \texttt{RSet} egy olyan típus, ami minden időpillanathoz egy másik típust rendel, ezután további segéd típusok következnek.\\

A \texttt{⇒} operátor segítségével reaktív típusok közti függvényképzést ad meg, konstans reaktív típusra pedig a \texttt{⟨A⟩} jelölést alkalmazza (ez minden időpillanathoz az \texttt{A} halmazt rendeli).

\begin{code}
  _⇒_ : RSet → RSet → RSet
  (A ⇒ B) t = A t → B t

  infixr 1 _⇒_

  ⟨_⟩ : Set → RSet
  ⟨ A ⟩ t = A

  ⟦_⟧ : RSet → Set
  ⟦ A ⟧ = ∀ {t} → A t
\end{code}

Fontos jelölés továbbá a \texttt{⟦ A ⟧} is, ami azon függvények típusa, amik minden t időponthoz hozzárendelnek egy \texttt{A t} halmazbeli elemet. Erre egy példa: \\

Legyenek t₁ és t₂ időpontok, amiken legyen definiálva egy rendezés:

\begin{code}
  postulate
    t₁ t₂ : Time
    _<_ : Time → Time → Bool
\end{code}

Legyen \texttt{A} egy reaktív (azaz időtől függő) típus, például legyen \texttt{A} t₁ és t₂ között \texttt{Bool}, egyébként pedig ℕ.

\begin{code}
  A : RSet
  A t with (t₁ < t) , (t < t₂)  -- A t
  ... | true , true = Bool      --  ∣ t₁ < t < t₂ = Bool
  ... | _ = ℕ                   --  ∣ egyébként   = ℕ
\end{code}

Ekkor \texttt{⟦ A ⟧} azon függvények típusa, amik minden t időponthoz hozzárendelnek egy \texttt{A t} beli értéket, azaz például egy függvény, aminek értéke t₁ és t₂ között false, egyébként pedig egy természetes szám:

\begin{code}
  f : ⟦ A ⟧
  f {t} with (t₁ < t) , (t < t₂)  -- f t
  ... | false , _     = 3         --  ∣      t < t₁ = 3
  ... | true  , true  = false     --  ∣ t₁ < t < t₂ = false
  ... | true  , false = 12        --  ∣ t₂ < t      = 12
\end{code}

A függvénykönyvtárban a további alacsony szintű típusok és függvények posztulátumként vannak megadva, ezekből a fordító számára külön megadott, kész ECMAScript kódokból generálódik a futtatható kód. Ennek jelölése egy $f$ függvényre:

\begin{center}
\texttt{\{\,-\# COMPILED\_JS $f$ <ECMAScript kód> \#-\,\}}
\end{center}

Ilyen típusok a \texttt{Beh} (Behaviour modul), \texttt{Evt} (Event modul), \texttt{DOM} és a hozzájuk tartozó függvények, a Behaviourök segítségével állapotátmenetek, Eventekkel az interakció valósítható meg, a \texttt{DOM} modulban pedig a HTML elemeket felépítő és módosító függvényeket definiálták, ezekből a következőket használjuk fel:

\begin{code}
  module Event where
   postulate
    Evt : RSet → RSet
    accumBy :  ∀ {A B} → ⟦ ⟨ B ⟩ ⇒ A ⇒ ⟨ B ⟩ ⟧ → B →
               ⟦ Evt A ⇒ Evt ⟨ B ⟩ ⟧

  module Behaviour where
   open Event
   postulate
    Beh : RSet → RSet
    map : ∀ {A B} → ⟦ A ⇒ B ⟧ → ⟦ Beh A ⇒ Beh B ⟧
    [_] : ∀ {A} → A → ⟦ Beh ⟨ A ⟩ ⟧
    hold : ∀ {A} → ⟦ ⟨ A ⟩ ⇒ Evt ⟨ A ⟩ ⇒ Beh ⟨ A ⟩ ⟧
\end{code}

A \texttt{[\_]} jelöléssel adhatunk meg konstans Behaviour értékeket, például ilyenek lesznek a programban a gombok feliratai és a HTML elemek stílusa. \\

Később, hogy ne kelljen a programban ilyen alacsony szintű típusokat és függvényeket használnunk, a Mastermind.View.Base modulban típusszinonimákkal elrejtjük ezeket a típusokat, így a program fő része áttekinthetőbb lesz. \\

Ezek után definiálták a Behaviour modulban a program működését biztosító, állapoton való léptetést biztosító függvényt, ezt accumHoldBy-nak nevezték el.

\begin{code}
   accumHoldBy :  ∀ {A B} → ⟦ (⟨ B ⟩ ⇒ A ⇒ ⟨ B ⟩) ⟧ → B →
                  ⟦ Evt A ⇒ Beh ⟨ B ⟩ ⟧
   accumHoldBy f b σ = hold b (accumBy f b σ)
\end{code}

Ennek első paramétere a léptetést megvalósító függvény, aminek szignatúrája \texttt{f : Allapot → Muvelet → Allapot}, második paramétere egy kezdeti állapot, a harmadik pedig a műveleteket tartalmazó esemény. A mi esetünkben az állapot egy rekord lesz, a műveletek pedig a gombok. \\

Az FRP.JS.DOM modulból a következő függvényeket használjuk majd:

\begin{itemize}
  \item element ,,<HTML tag>'' <elem>: statikus HTML elem
  \item text <szöveg>: szöveg hozzáadása
  \item attr ,,<HTML attribútum>'' <attr>: attribútum hozzáadása az aktuális HTML elemhez
\end{itemize}

Ezeknek a függvényeknek van <függvénynév>+ nevű változataik is, ezekkel lehet az interakciót megvalósító elemek (például gombok) tulajdonságait hasonlóan megadni.


\subsection{A program logikai és fizikai szerkezete}


\subsubsection{Mappaszerkezet}

A program mappaszerkezete a következő:

\begin{itemize}
  \item $dist$: A fordító által generált kimeneti fájlok mappája.
  \item $src$: A forrás, ezen belül az $agda$ mappa tartalmazza a program forrását, a $html$ pedig a HTML keretet és a stílust.
  \item $doc$: A dokumentáció.
\end{itemize}

\begin{figure}[H]
  \begin{center}
  \includegraphics{pic/files.pdf}
  \caption{Mappák hierarchiája}
  \end{center}
\end{figure}

A programhoz a gyökérkönyvtárban mellékeljük a $LICENSE$ fájlban a programhoz tartozó licencet, itt található továbbá a fordítást segítő $Makefile$ is, aminek használata a 2.4-es fejezetben olvasható.

\subsubsection{Modulok}

A programot a modell-nézet architektúra alapján építjük fel, azaz külön választjuk a megjelenést a
logikai szerkezettől. \\

\begin{figure}[H]
  \begin{center}
  \includegraphics{pic/hierarchy.pdf}
  \caption{Modul hierarchia}
  \end{center}
\end{figure}

A program kezdeti (\texttt{main}) függvénye a \texttt{Mastermind} modulban (\texttt{Mastermind.\allowbreak agda}) található, ezt a modult adjuk át a fordítónak. Ezután a modell-nézet architektúrának megfelelően különválasztjuk a megjelenést a logikai szerkezettől, a fő modul importálja a nézetet, amit a \texttt{Mastermind.View} (\texttt{Mastermind/View.agda}) modul implementál. A program logikai szerkezetét a \texttt{Mastermind.Model} (\texttt{Mastermind/Model.\allowbreak agda}) modulban helyeztük el, ebben találhatók az adatszerkezetek és az ezeket manipuláló függvények definíciói. \\

Az \texttt{agda-frp-js} függvénykönyvtár hátránya, hogy nem használnak az Agda standard library-ból semmilyen modult sem, és mi sem használhatjuk a kettőt együtt. Ennek hátránya az, hogy sok mindent nem vettek át a standard könyvtárból, ezért egy \texttt{Mastermind.Auxiliary} nevű modulban gyűjtöttem össze a hiányzó, de nekünk kellő függvényeket és típusokat. \\

A program tesztelését és a függvényekre vonatkozó állításokat a \texttt{Mastermind.\allowbreak Properties} (\texttt{Mastermind/Properties.agda}) valósítja meg, ez importálja a nézetet és a modellt is, őt pedig a fő modul (\texttt{Mastermind}) importálja, de nem használ fel belőle semmilyen függvényt. Ennek segítségével a tesztelés fordítás időben megtörté- nik, de ez nem változtatja meg a program működését. \\

A szakdolgozat dokumentációja a \texttt{Mastermind.Documentation} (\texttt{Mastermind/\allowbreak Documentation.lagda}) modulban található, ami Literate Agda kiterjesztésű, azaz helyes LaTeX és Agda kódot vegyesen tartalmaz, ebből például az $lhs2TeX$ nevű programmal tudunk $.tex$ fájlt előállítani, amiből pedig a $pdflatex$ parancs kiadásával elkészül a dokumentáció pdf formában. A dokumentációban található ábrák a $graphviz$ programmal készültek. \\

A megjelenítést HTML\cite{website:html4} (\texttt{main.html}) és CSS\cite{website:css} (\texttt{mastermind.css}) segítségével végezzük, ahol a \texttt{main.html} fájl betölti a lefordított ECMAScript kódokat egy keretbe, a stíluslap pedig a programban megadott HTML attribútumok (\texttt{class}, \texttt{id}) alapján színezi és igazítja az elemeket.


\newpage


\subsection{Modell}

\ignore{
\begin{code}
module Modell where
  open import FRP.JS.Nat hiding (_≟_ ; _<_)
  open import FRP.JS.List using (List ; _∷_ ; [] ; map ; _++_ ; length)
  open import FRP.JS.Bool using (Bool ; true ; false ; _∧_ ; if_then_else_ ; not) renaming (_≟_ to _≟b_)

  open import Mastermind.Auxiliary
\end{code}
}

A Modell (module Mastermind.Model) tartalmazza az állapot rekordot (record State), az összes állapotot módosító függvényt és a véletlenszámgenerátort (module Random).

\begin{figure}[H]
  \begin{center}
  \includegraphics{pic/model.pdf}
  \caption{Modell Diagram}
  \end{center}
\end{figure}

A modellben három konstans található, ezek a hagyományos Mastermind paramé- terei, ahol a minta 4 színből áll (codeLength), 12-őt tippelhetünk összesen (max- Guesses) és 6 színből választhatunk (colorCount).

\begin{code}
  codeLength = 4
  maxGuesses = 12
  colorCount = 6
\end{code}

\subsubsection{Típusok}

A modul néhány, pár elemű típust definiál, ezeket használva általánosabb típusok helyett jelentősen leszűkül az állapottér mérete és könnyebb az ezeket használó függvényekre állításokat is megfogalmazni. Ezek a típusok:

\begin{itemize}
  \item \texttt{Match}: black, white, none - találatok jelölése
  \item \texttt{End}: won, lost - játék végének esetei
  \item \texttt{GameState}: active, ended won, ended lost - játék állapota
  \item \texttt{Direction}: $\leftarrow$, $\rightarrow$ - irány jelölése
\end{itemize}

\ignore{
\begin{code}
  -- Találatok jelölése
  data Match : Set where
    black white none : Match

  -- Játék végének esetei
  data End : Set where
    won lost : End

  -- Játék állapota
  data GameState : Set where
    active : GameState
    ended : End → GameState

  data Direction : Set where
    ⟵ ⟶ : Direction
\end{code}
}

Ezek után típusszinonimákat adunk meg, amik csak a program olvashatóságát javítják, funkcionális szerepük nincs:

\ignore{
\begin{code}
  Color : Set
  Colors : Set
  Matches : Set
  HistElem : Set
  History : Set
\end{code}
}

\begin{code}
  Color     = ℕ
  Colors    = Vec Color codeLength
  Matches   = List Match
  HistElem  = Colors × Matches
  History   = List HistElem
\end{code}

A színek (\texttt{Color}) típusa természetes szám, így a backendre bízzuk a színek megjelenítését (\texttt{mastermind.css}), az ilyen színekből álló, kódhossz hosszúságú vektorok típusa a \texttt{Colors}. A találatok listájának szinonimája a \texttt{Matches}, a színek vektorából és ilyen találatok listájából álló pár típusa pedig a \texttt{HistElem}, a történet (\texttt{History}) ilyen elemek listája.

\subsubsection{Állapot}

Az állapotot rekordként adjuk meg, a következőképp:

\begin{code}
  record State : Set where
    field
      guess      : Colors     -- aktuális tipp
      solution   : Colors     -- megoldás
      history    : History    -- történet
      gamestate  : GameState  -- játék állapota
      rand       : ℕ          -- véletlen számok generálásához
                              -- szükséges kezdeti érték (seed)
      guesspos   : Fin.Fin 4  -- a tipp aktuálisan módosítható mezője
\end{code}

\begin{figure}[H]
  \begin{center}
  \includegraphics{pic/state.pdf}
  \caption{Állapot}
  \end{center}
\end{figure}

Ezen kívül megadunk állapot-módosító függvényeket, amiket majd a nézetből használunk, ezek:

\begin{itemize}
  \item \texttt{init}: kezdeti állapot, paramétere egy kezdeti érték a véletlenszám generátorhoz
  \item \texttt{incStateSeed}: a véletlenszám generátor léptetése, anélkül hogy felhasználnánk a generált értéket.
  \item \texttt{moveGuessPos}: a tipp aktuális mezőjén léptet a megadott irányba ($\leftarrow$, $\rightarrow$)
  \item \texttt{modifyColor}: az aktuális mezőn lévő szín változtatása
  \item \texttt{clearGuess}: kezdeti állapotba állítja a tippet (kék, kék, kék, kék)
  \item \texttt{newGame}: új játék kezdése
\end{itemize}

\subsubsection{Véletlenszámok}

Véletlenszámok generálásához szükséges egy modulo operátor, amit tisztán funkcio- nális nyelveken rekurzióval tudunk megadni, azonban itt a generálás során nagy számokkal használnánk, emiatt az Agda fordító a fordítás közbeni egyszerűsítések során nagy mértékben belassult, a generált ECMAScript kód pedig ,,too much recursion'' hibaüzenettel leállt. A lassulás arra vezethető vissza, hogy az Agda fordító megpróbálja a megadott konstansokkal elvégezni a modulo műveletet, ezzel egyszerűsítve a függvényhívást a véletlenszám generátorban. Ennek megakadályozására például az \texttt{abstract} kulcsszóval átláthatatlanná tehető a definíció, azonban ennek a használatával is fennállt a probléma. \\

A hiba kiküszöbölésére így a függvényt konstans 0-ként adjuk meg Agdában, a fordítónak pedig adunk egy hatékony JavaScript kódot használatra:

\begin{code}
  _%_ : ℕ -> ℕ -> ℕ
  n % m = 0
\end{code}

\texttt{\{\,-\# COMPILED\_JS \_\%\_ function(x) { return function(y) \\
 { if (x < 0) { return (x \% y) + x; } else { return x \% y; } }; } \#-\,\}} \\

Ennek hátránya, hogy így állításokat és teszteseteket nem tudunk helyesen megad- ni a véletlenszám generálással kapcsolatban. \\

\begin{figure}[H]
  \begin{center}
  \includegraphics{pic/random.pdf}
  \caption{Véletlenszámok}
  \end{center}
\end{figure}

Ezután következik a Random modul, ami egy paraméteres modul, paramétere a kezdeti érték (seed), mikor a program során használni szeretnénk, akkor az aktuális seed értékkel kell megnyitnunk a modult. Ez a modul egy úgynevezett Linear Congruential Generator (LCG) egy megvalósítása, Park és Miller ,,Minimal Standard''\cite{Press:1992:NRC:148286} változata alapján.

\begin{code}
  module Random (seed : ℕ) where
    private
      a = 16807
      m = 2147483647
      c = 0
\end{code}

Egy LCG három konstanst tartalmaz, ezek segítségével generál elemeket a követ- kező képlet alapján: $$X_{n+1} = (a * X_n + c) \bmod{m}$$ ahol $X_0 = seed$.

\begin{code}
      next : (n : ℕ) → ℕ
      next n = (((a * n) + c) % m)
\end{code}

\ignore{
\begin{code}
      gen-n : (n : ℕ) → ℕ → Vec ℕ n
      gen-n zero s = []
      gen-n (suc i) s = val ∷ gen-n i val where
        val = next s

      step-n : ℕ → ℕ
      step-n    0    = seed
      step-n (suc n) = next (step-n n)
\end{code}
}

Ezek segítségével a \texttt{newSolution} függvény generál egy új megoldást és visszaadja a megváltozott kezdeti értéket, amit aztán eltárolunk későbbi véletlen számok generálásához.

\begin{code}
    newSolution : Colors × ℕ
    newSolution = mapv (λ x → (x % colorCount) + 1) (gen-n 4 seed) , step-n 4
\end{code}

Itt a \texttt{gen-n} függvény \texttt{n} hosszú vektornyi véletlen számot állít elő, a \texttt{step-n} pedig \texttt{n}-et léptet a kezdeti értéken. A \texttt{newSolution} függvény mellett exportálunk egy \texttt{inc} nevű függvényt, amivel a program futása során majd minden gombnyomásra léptetünk egyet az aktuális seeden, így amikor új megoldás generálására van szükség, még véletlenebb értékeket kapunk.

\begin{code}
    inc : ℕ
    inc = step-n 1
\end{code}

Mivel az Agda nyelv tisztán funkcionális, így külső értéktől nem függhetnek a függvények, azonban az agda-frp-js függvénykönyvtárban a programok egy rejtett idő paramétertől is függnek, amit elérünk kezdetben, ez jól használható kezdeti értéknek.

\subsubsection{Tipp ellenőrzése}

A felhasználó által megadott tippet a \texttt{check} függvénnyel ellenőrizzük. Ennek működé- se a következő:

\begin{code}
  check : Colors → Colors → Matches
  check guess solution =
    blacks ++ whites
      where
        zipped = toList (zipv guess solution)

        diffs  = unzip (filter (not ∘ uncurry _≟_) zipped)
        rg     = proj₁ diffs
        rs     = proj₂ diffs
        whitec = length (rg \\ (rg \\ rs by _≟_) by _≟_)
        whites = replicate whitec white

        same   = filter (uncurry _≟_) zipped
        blacks = map (const black) same
\end{code}

Összefűzzük a tippet a megoldással (\texttt{zipv guess solution}), ezt listává alakítjuk (\texttt{toList}), így olyan listát kapunk (\texttt{zipped}) aminek minden eleme a tipp és a megoldás azonos pozíción lévő eleméből álló rendezett pár, ebből a fekete értékek (jó szín van jó helyen) száma egyenlő azon elemek számával, ahol a pár két eleme megegyezik (\texttt{same}). A fehér pöttyök (van olyan szín, de rossz helyen) esetén ezeket a mezőket eldobjuk (\texttt{diffs}), ezután a tippből maradt elemekből (\texttt{rg}) halmazként kivonva a megoldásból megmaradt elemeket (\texttt{rs}) megkapjuk a tipp azon elemeit, ami nem szerepelt a megoldásban. Ezután ezeket ismét halmazkivonással kivonva megkapjuk azokat az elemeket, amik szerepeltek, de rossz helyen, ezek száma (\texttt{whitec}) lesz a fehér pöttyök száma.


\newpage


\subsection{Nézet}

\ignore{
\begin{code}
module ViewBase where
  open import FRP.JS.Behaviour using ( Beh )
  open import FRP.JS.Product   using ( _∧_ )
  open import FRP.JS.Event     using ( Evt )
  open import FRP.JS.RSet      using ( ⟦_⟧ ; ⟨_⟩ ; _⇒_ )
  open import FRP.JS.DOM       using ( DOM )

  open import Mastermind.Auxiliary
  open import Mastermind.Model
\end{code}
}

A nézet tartalmazza a megjelenítéssel és az interakcióval kapcsolatos függvényeket. Ez a modul felhasználja a Modell által exportált állapot-módosító függvényeket és itt kezeljük a meg nem engedett eseteket is (pl. ha már vége a játéknak, akkor ne lehessen tippelni).

\begin{figure}[H]
  \begin{center}
  \includegraphics[width=0.8\textwidth]{pic/view.pdf}
  \caption{Megjelenítés}
  \end{center}
\end{figure}

Ez a modul felhasznál egy \texttt{Mastermind.View.Base} nevű modult is, ebben típusszinonimákkal elrejtem az agda-frp-js által megadott időtől függő típusokat, ettől a kód átláthatóbb lesz, és itt találhatók a nézet által felhasznált általánosabb függvények is. \\

A nézet egyetlen adatszerkezete a gombokat írja le (\texttt{Button}):

\begin{code}
  data Button : Set where
    color    : Color → Button      -- szín változtatása
    move     : Direction → Button  -- léptetés a tippben
    ok       : Button              -- tipp ellenőrzése
    clear    : Button              -- tipp törlése
    restart  : Button              -- játék újrakezdése
\end{code}

\ignore{
\begin{code}
  UIElement : Set
  Model : Set
  View : Set
  InteractiveButtons : Set
  ButtonHandler : Set
\end{code}
}

Ezután következnek a típusszinonimák:

\begin{code}
  InteractiveButtons  = ∀ {w} → ⟦ Beh (DOM w) ∧ Evt ⟨ Button ⟩ ⟧
  ButtonHandler       = Button → InteractiveButtons
\end{code}

Interaktív gombnak (\texttt{InteractiveButtons}) nevezzük el azt a típust, ami tartalmazza a gombok kinézetét (\texttt{Beh (DOM w)}) és a gombokkal történő interakciót (\texttt{Evt ⟨ Button ⟩}) is.

A \texttt{ButtonHandler} típus elemei olyan függvények, amik az általunk definiált gombokra ilyen interaktív gombokat hoznak létre. \\

Ezután következnek a megjelenítést végző függvények:

\begin{code}
  UIElement  = ∀ {w} → ⟦ Beh ⟨ State ⟩ ⇒ Beh (DOM w) ⟧
  Model      = ⟦ Evt ⟨ Button ⟩ ⇒ Beh ⟨ State ⟩ ⟧
  View       = ∀ {w} → ⟦ Beh (DOM w) ⟧
\end{code}

Az \texttt{UIElement} típus minden időpillanatra az aktuális állapotból készít egy megjeleníthető elemet (DOM), ezekből építjük fel az interfészt.

A \texttt{Model} gombokon értelmezett eseményekből állítja elő az állapot változását (\texttt{Behaviour}).

A \texttt{View} típus kész, kirajzolható HTML elemeket állít elő. \\

\ignore{
\begin{code}
module Nezet where

  open import FRP.JS.Behaviour using ( [_] ; accumHoldBy ; map ; join )
  open import FRP.JS.Product   using ( _,_ )
  open import FRP.JS.Event     using ( tag )
  open import FRP.JS.DOM       using ( element  ; attr    ; attr+ ; text
                                     ; text+    ; listen+ ; click ; _++_
                                     ; element+ ; _+++_   ) renaming ([] to []D)

  open import FRP.JS.Delay  using ( _ms )
  open import FRP.JS.Time   using ( Time ; epoch )

  open import FRP.JS.String using ( String ) renaming (_++_ to _++s_)
  open import FRP.JS.Maybe  using ( Maybe ; just ; nothing )
  open import FRP.JS.Nat    using ( ℕ ; zero ; suc )

  open import Mastermind.View.Base

  open import Mastermind.Model
  open import Mastermind.Auxiliary
\end{code}
}

A nézet egyik fő függvénye a \texttt{step}, ez valósítja meg az állapotban való lépést, típusa: \texttt{State → Button → State}, azaz megkapja az aktuális állapotot és a lenyomott gombot, és ezekből egy új állapotot állít elő. Ebben a függvényben ellenőrizzük továbbá, hogy az adott gomb használható volt-e, mikor lenyomták (például ha vége a játéknak, akkor nem lehet többet tippelni), ez azért szükséges itt, mert a nézetben definiáltuk a gombokat is. \\

\ignore{
\begin{code}
  step : State → Button → State
  step state restart = newGame state
  step state button with State.gamestate state
  ... | ended _ = state
  ... | active with button
  ...    |    color c     = modifyColor c state
  ...    |       ok       = checkGuess state
  ...    |     clear      = clearGuess state
  ...    | move direction = moveGuessPos direction state
  ...    |       _        = state
\end{code}
}

Ezután megadjuk egy \texttt{model} nevű függvényt, ami a felhasznált módszereknél ismertetett \texttt{accumHoldBy} függvényt felhasználva, a modell \texttt{init} függvényével előállít egy kezdeti állapotot, amit aztán a \texttt{step} függvénnyel léptet tovább. Ez egy reaktív típusú függvény, így elő tudjuk hozni a rejtett paraméterét (\texttt{epoch (<idő> ms)}), ami a program indításának időpontja, ezt adjuk át az \texttt{init} függvénynek, azaz ez lesz a véletlenszám generátor kezdeti értéke.

\begin{code}
  model : Model
  model {epoch (time ms)} =
    accumHoldBy step (init time)
\end{code}

Interaktív gombokat a \texttt{button} függvénnyel hozunk létre, ami adott gombra előállítja az eseményt, a címkéjét és a stílusát.

\begin{code}
  button : ButtonHandler
  button button =
    listen+ click (λ _ → button) (              -- esemény
      element+ "button" (                       -- HTML gomb
        text+ [ button$ button ] +++            -- felirat
        attr+ "class" [ buttonClass button ]))  -- HTML osztály
\end{code}

Az összes használható gombot a \texttt{keypad : InteractiveButtons} állítja elő, ez a függvény adja meg a sorrendjüket és az elhelyezkedésüket is. \\

Ezek után következik a nézet nem interaktív része, az itt lévő típusok az állapotot képzik le HTML elemekre, eseményeket nem állítanak elő, így típusukban a visszatérési érték a fentebb leírt \texttt{UIElement} típusú. \\

A \texttt{displayGuess : UIElement} függvény a tippet jeleníti meg, a \texttt{display- History : UIElement} pedig a teljes történetet, ennek mind a 12 sora 2 részből áll, az akkori tippből, és a rá kapott fekete/fehér pöttyökből. \\

\ignore{
\begin{code}
  postulate
    displayGuess : UIElement
    displayHistory : ℕ → UIElement
    keypad : InteractiveButtons
\end{code}
}

A \texttt{display} függvény használatával állítjuk elő a nézet teljes nem interaktív részét, a történetből táblázatot képzünk, alatta helyezkedik el az aktuális tipp, utána pedig, ha véget ért a játék egy felirat lesz látható.

\begin{code}
  display : UIElement
  display σ =
    element "table" (                    -- történet
      displayHistory maxGuesses σ ++
      attr "id" [ "historytable" ] ) ++
    element "div" (displayGuess σ) ++    -- aktuális tipp
    element "div" (text (map state$ σ))  -- nyert/vesztett felirat
\end{code}

A teljes nézetet a \texttt{view} függvény állítja elő, ez létrehozza a nézetet az interaktív gombok eseményeivel (\texttt{display (model evt)}), utána pedig hozzáfűzi a gombok HTML elemeit. Ezt a függvényt használja fel a fő (\texttt{Mastermind}) modulban lévő belépési pont (\texttt{main}).

\begin{code}
  view : View
  view with keypad
  ... | (dom , evt) = display (model evt) ++ dom
\end{code}

\subsection{HTML és CSS}

A lefordított kód betöltését a \texttt{main.html} nevű HTML fájl végzi, ebben egy egyszerű táblázat segítségével egy középre igazított keretbe töltődik be a program, ennek a kódja:

\begin{center}
\texttt{<div class="agda" id="mastermind" data-agda="Mastermind"></div>}
\end{center}

A stíluslap (CSS) a (\texttt{mastermind.css}) fájlban található, ami többféle osztályt és azonosítót is használ, ezek a következők:

\begin{itemize}
  \item Általános:
    \begin{itemize}
      \item \texttt{.agda}       - a teljes programra vonatkozó stílus
      \item \texttt{\#maintable} - a teljes képernyőt kitevő keret
      \item \texttt{\#maintitle} - ,,Mastermind'' felirat kinézete
      \item \texttt{\#mainpart}  - a program kerete
    \end{itemize}
  \item Korábbi próbálkozásokra vonatkozó:
    \begin{itemize}
      \item \texttt{\#historytable}  - a keret stílusa
      \item \texttt{\#matchtable}    - a találatok kerete
      \item \texttt{\#hist\_matches} - a találatok mérete
      \item \texttt{\#black\_m}, \texttt{\#white\_m} - a találatok színe
    \end{itemize}
  \item Aktuális tippre vonatkozó:
    \begin{itemize}
      \item \texttt{\#guesstable} - a keret stílusa
      \item \texttt{.color}       - minden színre vonatkozó értékek
      \item \texttt{.color[1-6]}  - színek hozzárendelése a programban használt reprezentációhoz (ℕ)
      \item \texttt{\#guess\_act} - a tipp aktuális elemének megkülönböztetésére használt stílus
    \end{itemize}
  \item Gombok:
    \begin{itemize}
      \item \texttt{.button\_num} - színek gombjai
      \item \texttt{.button\_op}  - műveletek gombjai
    \end{itemize}
\end{itemize}


\subsection{Tesztelési terv}

\ignore{
\begin{code}
module Testing where
  open import FRP.JS.Bool renaming (_≟_ to _≟b_)
  open import FRP.JS.List
  open import FRP.JS.Maybe
  open import FRP.JS.Nat hiding (_≟_ ; _<_)

  open import Mastermind.Auxiliary

  open Mastermind.Auxiliary.Fin

  open import Mastermind.Model
  open import Mastermind.View
  open import Mastermind.View.Base
\end{code}
}

A szoftver teszteléséhez az Agda nyelv beépített tételbizonyító rendszerét használjuk fel, amivel a program futása során fennálló összefüggések mellett tesztesetek is megadhatók. Ezeket a \texttt{Mastermind.Properties} modulban definiáltuk, amit a leg- felső (\texttt{Mastermind}) modul importál, így a tesztelés fordítás közben zajlik, ezzel garantálva, hogy csak helyesen működő programot lehessen lefordítani.

\subsubsection{Nyelvi elemek}

Az Agda nyelv szigorú típusrendszere és különféle nyelvi elemei segítségével elkerülhetjük a hibás állapotba lépést is már a fordítás során, ami a tesztelést is megkönnyíti. Ezek a nyelvi elemek minden függvényre/típusra adottak:

\begin{itemize}
  \item \emph{Totális függvények:}
    A függvény az állapotterének minden elemét le kell hogy képezze valamilyen elemre.
  \item \emph{Terminálás ellenőrzés:}
    A függvények nem kerülhetnek végtelen ciklusba (minden rekurziónak struktálisan
    kisebbnek kell lennie az eredeti függvényhívásnál, stb.).
\end{itemize}

A teszteléshez felhasználjuk a propozícionális egyenlőséget (Propositional Equality) \cite{website:abel:equality}, amit a \texttt{Mastermind.Properties} modul elején definiálunk.

\begin{code}
  data _≡_ {A : Set} (a : A) : A → Set where
    refl : a ≡ a

  infix 4 _≡_

  data ⊥ : Set where

  infix 3 ¬_

  ¬_ : ∀ {ℓ} → Set ℓ → Set ℓ
  ¬ P = P → ⊥

  _≢_ : ∀ {A : Set} → A → A → Set
  x ≢ y = ¬ x ≡ y
\end{code}

\texttt{∀ (a, b : A) : a ≡ b} egy olyan két paraméteres adatszerkezet, aminek akkor van eleme, ha a két paramétere egyenlő, ez az elem pedig a \texttt{refl : a ≡ a}, azaz a reflexió axióma szerint definiáltuk. Az ekvivalencia tagadása (\texttt{≢}) az üres halmaz (\texttt{⊥}) bevezetésével valósítható meg, ahol a tagadás az üres halmazra való képzés (\texttt{¬ P = P → ⊥}). \\

Mivel a propozícionális egyenlőség egy ekvivalenciareláció, így a szimmetria (\texttt{sym}) és a tranzitivitás (\texttt{trans}) is teljesül rá, ezeket szintén definiáljuk.

\begin{code}
  sym : ∀ {A : Set} {a b : A} → (a ≡ b) → (b ≡ a)
  sym refl = refl

  trans : ∀ {A : Set} {a b c : A} → (a ≡ b) → (b ≡ c) → (a ≡ c)
  trans refl refl = refl
\end{code}

\subsubsection{Tesztelés}

A teszteléshez felhasználjuk a nézetből a \texttt{step} függvényt, a modellből pedig az állapotot változtató függvényeket, ezek segítségével tetszőleges állapotból eljuthatunk azokba az állapotokba, amikbe a felhasználó is. Példaként tetszőleges állapotból az újrakezdés (\texttt{restart}) gombot lenyomva új játék kezdődik. Ennek bizonyítása a definíciókból következik:

\begin{code}
  restart≡newGame : ∀ {s : State} → step s restart ≡ newGame s
  restart≡newGame = refl
\end{code}

Emellett megadunk egy további függvényt (\texttt{>>}) a lépések sorozatának jelölésére a könnyebb olvashatóság segítéséhez.

\begin{code}
  _>>_ : State → Button → State
  state >> button = step state button

  infixl 10 _>>_
\end{code}

Ezzel könnyen felírható az első teszteset (\texttt{test₁}), ami szerint kezdőállapotból (\texttt{init}) 4 szín megadása (\texttt{color x}) után az állapotban lévő tipp a megadott 4 szín lesz.

\begin{code}
  test₁ : {n : ℕ} {a b c d : Color} → 
    State.guess (init n >> color a >> color b >> color c >> color d)
    ≡
    (a ∷ b ∷ c ∷ d ∷ [])
  test₁ = refl
\end{code}

A következő tesztesetnél a tippben való léptetésre szeretnénk belátni, hogy nem léphet ki a tippből, azaz a \texttt{State.guesspos} értéke nem lehet 4-nél nagyobb. Mivel ennek az értéknek a típusa \texttt{Fin 4}, azaz 4 elemű véges halmaz, így erre a típusra kell egy egyszerű lemmát megadnunk. Ez a lemma azt mondja ki, hogy \texttt{n} elemű halmaz tetszőleges \texttt{i} indexű eleménél \texttt{i < n} mindig teljesül.

\begin{code}
  lemma-fin : ∀ {n} → (i : Fin n) → toℕ i < n ≡ true
  lemma-fin zero    = refl
  lemma-fin (suc i) = lemma-fin i
\end{code}

Ezután a mi esetünk ennek a lemmának egy speciális esete, ahol \texttt{n = 4} és \texttt{i = State.guesspos s}, ehhez fel is használjuk az előbb megadott függvényt.

\begin{code}
  test₂ : (s : State) → toℕ (State.guesspos s) < 4 ≡ true
  test₂ s = lemma-fin (State.guesspos s)
\end{code}

A tesztesetek kiterjedhetnek több gomb működésére is, például beláthatjuk, hogy ha véget ért a játék és nem a 'restart' gombot nyomtuk meg, akkor nem változik az állapot (\texttt{step s b ≡ s}) a következő típusú függvény megadásával:

\begin{code}
  no-step-when-ended :
    ∀ {e : End} → (s : State) → (b : Button) →
    b ≢ restart → ‌State.gamestate s ≡ ended e →
    step s b ≡ s
\end{code}

\ignore{
\begin{code}
  no-step-when-ended s restart np p with np refl
  no-step-when-ended s restart np p | ()
  no-step-when-ended s (color y) np p  with State.gamestate s
  no-step-when-ended s (color y) np () | active
  no-step-when-ended s (color y) np refl | ended _ = refl
  no-step-when-ended s (move y) np p  with State.gamestate s
  no-step-when-ended s (move y) np () | active
  no-step-when-ended s (move y) np refl | ended _ = refl
  no-step-when-ended s ok np p with State.gamestate s
  no-step-when-ended s ok np () | active
  no-step-when-ended s ok np refl | ended _ = refl
  no-step-when-ended s clear np p with State.gamestate s
  no-step-when-ended s clear np () | active
  no-step-when-ended s clear np refl | ended _ = refl

  lemma-lookupv :
    ∀ {n : ℕ} {A : Set} {a : A}
    (fn : Fin n) (v : Vec A n) →
    lookupv (replacev (toℕ fn) v a) (toℕ fn) ≡ just a
  lemma-lookupv () []
  lemma-lookupv zero (x ∷ xs) = refl
  lemma-lookupv (suc i) (x ∷ xs) = lemma-lookupv i xs
\end{code}
}

Az állításokban hivatkozhatunk az állapot több komponensére is, például szeretnénk belátni, hogy ha egy állapotban (\texttt{s}) megváltoztatjuk az aktuális pozíción lévő színt (\texttt{modifyColor}), akkor az új állapotban (\texttt{modifyColor c s}) az eredeti tipp aktuális pozícióján (\texttt{State.guesspos s}) lévő szín a megadott szín lesz:

\begin{code}
  test-modifyColor : 
    ∀ {c : Color} (s : State) →
    lookupv (State.guess (modifyColor c s))  -- megváltozott állapot tippjének
            (toℕ (State.guesspos s))         -- az eredeti állapot tippjének
                                             -- aktuális pozícióján
    ≡ just c                                 -- az új szín lesz
  test-modifyColor s = lemma-lookupv (State.guesspos s) (State.guess s)
\end{code}

\subsubsection{Állítások}

Az interfész tesztelése mellett a modell további függvényeire is adhatunk meg állításokat, például szeretnénk belátni, hogy ha a tippünk megegyezik a megoldással, akkor eredményként 4 fekete pöttyöt kapunk, ami megoldása is a játéknak. Ehhez először bebizonyítjuk, hogy a 4 fekete pötty az megoldás, ez mivel egyszerű összehasonlítás, definícióból következik:

\begin{code}
  isSolved-law₁ : isSolved (black ∷ black ∷ black ∷ black ∷ []) ≡ true
  isSolved-law₁ = refl
\end{code}

Ezután a \texttt{check} függvényről szeretnénk belátni, hogy ha a tipp és a megoldás paramétere megegyezik, akkor eredményül 4 fekete pöttyöt kapunk. Ennek bizonyítása nem triviális, mivel maga a függvény is egy összetettebb algoritmust használ. \\

Ilyen, bonyolultabb állítások bizonyításához felhasználhatjuk az egyenlőségi érve- lést, ami egy speciálisan zárójelezett infix operátor, a tranzitivitást felhasználva több részre bontja a bizonyítást.

\begin{code}
  _≡⟨_⟩_ : ∀ {A : Set} (x : A) {y z : A} → x ≡ y → y ≡ z → x ≡ z
  x ≡⟨ refl ⟩ refl = refl

  infixr 2 _≡⟨_⟩_

  _∎ : ∀ {A : Set} (x : A) → x ≡ x
  x ∎ = refl

  infix 2 _∎
\end{code}

Bizonyítások során használt fogalom továbbá a kongruencia (\texttt{cong}), ami azt fejezi ki, hogy ha egy \texttt{a₁} és \texttt{a₂} azonos típusbeli értékek megegyeznek, akkor egy rajtuk elvégzett \texttt{f} függvény eredménye is megegyezik:

\begin{code}
  cong :  ∀ {A : Set} {B : Set} (f : A → B) {a₁ a₂ : A} →
          (a₁ ≡ a₂) → (f a₁ ≡ f a₂)
  cong f refl = refl
\end{code}

Ennek segítségével a fentebb leírt bizonyításban először belátjuk, hogy a fehér találatok száma 0, így az őket tartalmazó lista üres (\texttt{law₁ : whites ≡ [ ]}), ezután levezetjük, hogy a fekete találatokból álló lista 4 fekete elemet tartalmaz (\texttt{law₂ : blacks ≡ (black ∷ black ∷ black ∷ black ∷ [ ])}). Ezek- nek az állításoknak a bizonyítására további állításokra és lemmákra van szükség, amiket itt nem részletezünk.

\begin{code}
  check-law₁ : (c : Colors) → check c c ≡ (black ∷ black ∷ black ∷ black ∷ [])
  check-law₁ c =
    check c c
      ≡⟨ refl ⟩                           -- check definíciója
    blacks ++ whites
      ≡⟨ cong (λ x → blacks ++ x) law₁ ⟩  -- law₁ felhasználása
    blacks ++ []
      ≡⟨ cong (λ x → x ++ []) law₂ ⟩      -- law₂ felhasználása
    black ∷ black ∷ black ∷ black ∷ []
      ∎                                   -- QED
\end{code}

\begin{code}
    where
\end{code}

\ignore{
\begin{code}
      zipped = toList (zipv c c)

      rr     = unzip (filter (not ∘ uncurry _≟_) zipped)
      rg     = proj₁ rr
      rs     = proj₂ rr
      whitec = length (rg \\ (rg \\ rs by _≟_) by _≟_)
      whites = replicate whitec white

      fs     = filter (uncurry _≟_) zipped
      blacks = map (λ _ → black) fs
      postulate
\end{code}
}

\begin{code}
       law₁ : whites ≡ []
       law₂ : blacks ≡ (black ∷ black ∷ black ∷ black ∷ [])
\end{code}

Ezután ezt a bizonyítást felhasználva egyszerűen megadhatjuk, hogy megegyező tipp és megoldás esetén a játék megoldását kapjuk:

\begin{code}
  check-law₁-solution : ∀ (c : Colors) → isSolved (check c c) ≡ true
  check-law₁-solution c =
    isSolved (check c c)
      ≡⟨ cong isSolved (check-law₁ c) ⟩
    isSolved (black ∷ black ∷ black ∷ black ∷ [])
      ≡⟨ refl ⟩
    true
      ∎
\end{code}


\newpage


\section{Összegzés}


A szakdolgozat elkészítésével sikerült az Agda, egy új és főleg matematikai célokra használt programozási nyelv segítségével egy grafikus felületű, többféle platformon is futtatható programot megvalósítani. \\

Ennek az implementációnak a legfőbb előnye az, hogy a tesztelést a program kódjának fordításával együtt végzi el a fordító, így futtatható állomány kizárólag a teszteseteknek megfelelő működés esetén jön létre, továbbá nem volt szükség külső tesztelő szoftverek használatára sem, ezzel felgyorsítva a fejlesztés és tesztelés folyamatát. \\

A funkcionális programok futtatása esetén fellépő hatékonysági problémákat, így például a futás közbeni véletlenszám generálást előre fordított JavaScript függvények használatával sikerült elkerülni, így a mai elvárásoknak megfelelő sebességű program állt elő. \\

További fejlesztési lehetőségek a programban:

\begin{itemize}
  \item A Mastermind játék általánosítása, tetszőleges számú tippelési lehetőséggel és tipp hosszúsággal.
  \item Toplista megvalósítása, azaz I/O műveletek és az agda-frp-js függvénykönyvtár együttes használata, vagy a függvénykönyvtár kiterjesztése
\end{itemize}


\newpage


\bibliography{bibliography.bib}
\bibliographystyle{plain}


\end{document}
